#FILE: get_snps_by_gene_ontology.pl
#AUTH: Paul Stothard (paul.stothard@gmail.com)
#DATE: March 6, 2010
#VERS: 1.0

use warnings;
use strict;
use Data::Dumper;
use Bio::EnsEMBL::Registry;
use Bio::Seq;
use Tie::IxHash;
use Getopt::Long;

my %options = (
    go_accession => undef,
    species      => undef,
    flanking     => 10,
    output       => undef
);

GetOptions(
    'g=s' => \$options{go_accession},
    's=s' => \$options{species},
    'f=i' => \$options{flanking},
    'o=s' => \$options{output}
);

if (   !( defined( $options{go_accession} ) )
    or !( defined( $options{output} ) )
    or !( defined( $options{species} ) ) )
{
    die( print_usage() . "\n" );
}

my $registry = 'Bio::EnsEMBL::Registry';

$registry->load_registry_from_db(
    -host => 'ensembldb.ensembl.org',
    -user => 'anonymous',
    -pass => ''
);

my $gene_adaptor
    = $registry->get_adaptor( $options{species}, 'Core', 'Gene' );

my $go_adaptor = $registry->get_adaptor( 'Multi', 'Ontology', 'GOTerm' );

my $term = $go_adaptor->fetch_by_accession( $options{go_accession} );

my $trv_adaptor = $registry->get_adaptor( $options{species}, 'variation',
    'transcriptvariation' );

my $va_adaptor
    = $registry->get_adaptor( $options{species}, 'variation', 'variation' );

my $first_record = 1;

my @column_titles = ();

open( my $OUTFILE, '>', $options{output} )
    or die("Cannot open file '$options{output}': $!");

print
    "Retrieving $options{species} genes associated with the following GO term:\n";
print "GO ID: " . $options{go_accession} . "\n";
print "GO name: " . $term->name() . "\n";
print "GO definition: " . $term->definition() . "\n";

my $record_count = 0;
foreach my $gene ( @{ $gene_adaptor->fetch_all_by_GOTerm($term) } ) {

    print "Obtaining SNPs for the following gene:\n";
    print "Ensembl Gene ID: " . $gene->stable_id . "\n";

    my $transcript = $gene->canonical_transcript();

    foreach my $tv (
        @{ $trv_adaptor->fetch_all_by_Transcripts( [$transcript] ) } )
    {

        tie( my %output, 'Tie::IxHash' );

        #order here determines order in output
        #key names determine column labels
        $output{entrez_gene_name} = undef;
        $output{entrez_gene_id}   = undef;
        $output{transcript_id}    = undef;
        $output{gene_id}          = undef;
        $output{gene_description} = undef;
        $output{gene_ontology}    = undef;
        $output{snp_id}           = undef;
        $output{type}             = undef;
        $output{chromosome}       = undef;
        $output{strand}           = undef;
        $output{start}            = undef;
        $output{end}              = undef;
        $output{sequence}         = undef;

        #Entrez Gene name and Entrez Gene ID
        my $xrefs = $transcript->get_all_DBLinks();
        foreach my $xref ( @{$xrefs} ) {
            if ( $xref->dbname() eq 'EntrezGene' ) {
                $output{entrez_gene_name} = $xref->display_id();
                $output{entrez_gene_id}   = $xref->primary_id();
            }
        }

        $output{transcript_id}    = $transcript->stable_id;
        $output{gene_id}          = $gene->stable_id;
        $output{gene_description} = $gene->description();

        #Gene ontology
        my @go_names = ();
        foreach my $xref ( @{$xrefs} ) {
            my $go_id         = $xref->display_id();
            my $go_term       = $go_adaptor->fetch_by_accession($go_id);
            my $go_name       = $go_term->name();
            my $go_definition = $go_term->definition();

            if ( defined($go_name) ) {
                push( @go_names, $go_name );
            }
        }
        $output{gene_ontology} = join( "; ", @go_names );

        $output{snp_id}     = $tv->variation_feature->variation_name;
        $output{type}       = join( ";", @{ $tv->consequence_type } );
        $output{chromosome} = $tv->variation_feature->seq_region_name;
        $output{strand}     = $tv->variation_feature->seq_region_strand;
        $output{start}      = $tv->variation_feature->seq_region_start;
        $output{end}        = $tv->variation_feature->seq_region_end;

        my $var = $va_adaptor->fetch_by_name( $output{snp_id} );

        $output{sequence} = substr( $var->five_prime_flanking_seq,
            ( -1 * $options{flanking} ) )
            . '['
            . $tv->variation_feature->allele_string() . ']'
            . substr( $var->three_prime_flanking_seq, 0, $options{flanking} );

        #prepare output array
        my @output = ();
        my @keys   = keys(%output);
        foreach my $key (@keys) {
            if (   ( !defined( $output{$key} ) )
                || ( !( $output{$key} =~ m/\S/ ) ) )
            {
                $output{$key} = '?';
            }

            push( @output, $output{$key} );

            if ($first_record) {
                push( @column_titles, $key );
            }
        }

        #clean up @output
        map( s/[\n\t\'\"]/ /g, @output );
        map( s/\s{2,}/ /g,     @output );
        map( s/^\s+//g,        @output );
        map( s/\s+$//g,        @output );

        if ($first_record) {
            $first_record = 0;
            print $OUTFILE join( "\t", @column_titles ) . "\n";
        }

        print $OUTFILE join( "\t", @output ) . "\n";
        $record_count++;

        print "SNP record $record_count written to $options{output}\n";
    }
}

close($OUTFILE) or die("Cannot close file : $!");

print "$record_count SNPs written to $options{output}\n";

sub print_usage {
    print <<BLOCK;
USAGE: perl get_snps_by_gene_ontology.pl [-arguments]
 -s [STRING]  : the source species for the SNPs (Required).
 -g [STRING]  : the gene ontology ID of interest (Required).
 -o [FILE]    : the output file to create (Required).
 -f [INTEGER] : amount of flanking genomic sequence to obtain on
                either side of SNP (Optional. Default is 10 bases).

perl get_snps_by_gene_ontology.pl -s bos_taurus -g 'GO:0030326' \
-f 250 -o results.tab
BLOCK
}
